import re
import inspect

from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist
from django.db.models import Model
from django.db.models.fields import FieldDoesNotExist
from django.forms.forms import pretty_name
from django.utils.encoding import force_text
from django.utils.translation import ugettext
from django.utils.html import format_html

from chamber.utils import get_class_method, call_method_with_unknown_input

from pyston.converters import get_converter


def is_callable(val):
    return hasattr(val, '__call__')


class FieldOrMethodDoesNotExist(Exception):
    pass


class InvalidMethodArguments(Exception):
    pass


def str_to_class(class_string):
    module_name, class_name = class_string.rsplit('.', 1)
    # load the module, will raise ImportError if module cannot be loaded
    m = __import__(module_name, globals(), locals(), str(class_name))
    # get the class, will raise AttributeError if class cannot be found
    c = getattr(m, class_name)
    return c


def get_new_class_name(prefix, klass):
    prefix = prefix.replace('-', ' ').title()
    prefix = re.sub(r'\s+', '', prefix)
    return prefix + klass.__name__


def flatten_fieldsets(fieldsets):
    """Returns a list of field names from an admin fieldsets structure."""
    field_names = []
    for _, opts in fieldsets or ():
        if 'fieldsets' in opts:
            field_names += flatten_fieldsets(opts.get('fieldsets'))
        else:
            for field in opts.get('fields', ()):
                if isinstance(field, (list, tuple)):
                    field_names.extend(field)
                else:
                    field_names.append(field)
    return field_names


def get_inline_views_from_fieldsets(fieldsets):
    """Returns a list of field names from an admin fieldsets structure."""
    inline_views = []
    for _, opts in fieldsets or ():
        if 'fieldsets' in opts:
            inline_views += get_inline_views_from_fieldsets(opts.get('fieldsets'))
        elif 'inline_view_inst' in opts:
            inline_views.append(opts.get('inline_view_inst'))
    return inline_views


def get_inline_views_opts_from_fieldsets(fieldsets):
    """Returns a list of field names from an admin fieldsets structure."""
    inline_views = []
    for _, opts in fieldsets or ():
        if 'fieldsets' in opts:
            inline_views += get_inline_views_opts_from_fieldsets(opts.get('fieldsets'))
        elif 'inline_view' in opts:
            inline_views.append(opts)
    return inline_views


def get_field_from_model_or_none(model, field_name):
    try:
        return model._meta.get_field(field_name)
    except (FieldDoesNotExist, AttributeError):
        return None


def _get_verbose_value(raw, field_or_method, obj, **kwargs):
    if hasattr(field_or_method, 'humanized') and field_or_method.humanized:
        return field_or_method.humanized(raw, obj, **kwargs)
    elif hasattr(field_or_method, 'attname') and hasattr(obj, 'get_{}_display'.format(field_or_method.attname)):
        return getattr(obj, 'get_{}_display'.format(field_or_method.attname))()
    else:
        return raw


def _val_to_readonly_value(value, field_or_method, obj, **kwargs):
    from is_core.forms.utils import ReadonlyValue
    verbose_value = _get_verbose_value(value, field_or_method, obj, **kwargs)
    return ReadonlyValue(value, verbose_value) if value != verbose_value else value


def _get_method_value(method, field_name, inst, fun_kwargs):
    method_kwargs_names = inspect.getargspec(method)[0][1:]
    method_kwargs = {arg_name: fun_kwargs[arg_name]  for arg_name in method_kwargs_names if arg_name in fun_kwargs}
    if len(method_kwargs_names) == len(method_kwargs):
        return _val_to_readonly_value(method(**method_kwargs), method, inst,
                                     **{k: v for k, v in method_kwargs.items() if k != 'obj'})
    else:
        raise(InvalidMethodArguments('Method {} arguments has not subset of fun kwargs'.format(field_name)))


def _get_method_or_property_value(class_method, field_name, inst, fun_kwargs):
    method_or_property = getattr(inst, field_name)

    return (
        _get_method_value(method_or_property, field_name, inst, fun_kwargs) if is_callable(method_or_property)
        else _val_to_readonly_value(method_or_property, class_method, inst)
    )


def get_model_method_or_property_data(field_name, model, fun_kwargs):
    from is_core.forms.widgets import ReadonlyWidget

    class_method = get_class_method(model, field_name)

    method = (
        getattr(model, field_name)
        if hasattr(model, field_name) and not class_method and is_callable(getattr(model, field_name))
        else class_method
    )

    if method:
        if hasattr(method, 'field'):
            # Generic relation
            label = getattr(method.field, 'verbose_name', pretty_name(field_name))
        else:
            label = getattr(method, 'short_description', pretty_name(field_name))
        try:
            return (
                (None, label, ReadonlyWidget) if isinstance(model, type)
                else (_get_method_or_property_value(method, field_name, model, fun_kwargs), label, ReadonlyWidget)
            )
        except InvalidMethodArguments:
            return None
    elif hasattr(model, field_name):
        return (
            getattr(model, field_name), pretty_name(field_name), ReadonlyWidget
        )
    else:
        return None


def _get_model_field_data(field, model):
    from is_core.forms.widgets import ReadonlyWidget, ManyToManyReadonlyWidget, ModelObjectReadonlyWidget

    if field.auto_created and (field.one_to_many or field.many_to_many):
        return (
             None if isinstance(model, type) else [obj for obj in getattr(model, field.name).all()], (
                getattr(field.field, 'reverse_verbose_name', None)
                if getattr(field.field, 'reverse_verbose_name', None) is not None
                else field.related_model._meta.verbose_name_plural
            ), ManyToManyReadonlyWidget
        )
    elif field.auto_created and field.one_to_one:
        return (
            None if isinstance(model, type) or not hasattr(model, field.name)
            else getattr(model, field.name), (
                getattr(field.field, 'reverse_verbose_name', None)
                if getattr(field.field, 'reverse_verbose_name', None) is not None
                else field.related_model._meta.verbose_name
            ),  ModelObjectReadonlyWidget
        )
    elif field.many_to_many:
        return (
            None if isinstance(model, type) else [obj for obj in getattr(model, field.name).all()],
            field.verbose_name, ManyToManyReadonlyWidget
        )
    else:
        return (
            None if isinstance(model, type)
            else _val_to_readonly_value(
                getattr(model, field.name) if hasattr(model, field.name) else None, field, model
            ), field.verbose_name, ReadonlyWidget
        )


def _get_model_readonly_data(field_name, model, fun_kwargs):
    if '__' in field_name:
        current_field_name, next_field_name = field_name.split('__', 1)
        field = get_field_from_model_or_none(model, current_field_name)
        if hasattr(model, current_field_name) and getattr(model, current_field_name):
            return _get_model_readonly_data(next_field_name, getattr(model, current_field_name), fun_kwargs)
        elif model and field and hasattr(field, 'related_model'):
            return _get_model_readonly_data(next_field_name, field.related_model, fun_kwargs)
        else:
            return None
    else:
        field = get_field_from_model_or_none(model, field_name)
        return (
            _get_model_field_data(field, model) if field
            else get_model_method_or_property_data(field_name, model, fun_kwargs)
        )


def _get_view_readonly_data(field_name, view, fun_kwargs):
    from is_core.forms.widgets import ReadonlyWidget

    method_field = view.get_method_returning_field_value(field_name)
    if method_field:
        try:
            return (
                _get_method_value(method_field, field_name, view, fun_kwargs),
                getattr(method_field, 'short_description', pretty_name(field_name)),
                ReadonlyWidget
            )
        except InvalidMethodArguments:
            return None
    else:
        return None


def get_readonly_field_data(field_name, instance, view=None, fun_kwargs=None):
    """
    Returns field humanized value, label and widget which are used to display of instance or view readonly data.
    Args:
        field_name: name of the field which will be displayed
        instance: model instance
        view: view instance
        fun_kwargs: kwargs that can be used inside method call

    Returns:
        field humanized value, label and widget which are used to display readonly data
    """
    fun_kwargs = fun_kwargs or {}

    if view:
        view_readonly_data = _get_view_readonly_data(field_name, view, fun_kwargs)
        if view_readonly_data is not None:
            return view_readonly_data

    field_data = _get_model_readonly_data(field_name, instance, fun_kwargs)
    if field_data is not None:
        return field_data

    raise FieldOrMethodDoesNotExist('Field or method with name {} not found'.format(field_name))


def display_object_data(obj, field_name, request=None):
    """
    Returns humanized value of model object that can be rendered to HTML or returned as part of REST

    examples:
       boolean True/Talse ==> Yes/No
       objects ==> object display name with link if current user has permissions to see the object
       field with choices ==> string value of choice
       field with humanize function ==> result of humanize function
    """
    from is_core.forms.utils import ReadonlyValue

    value, _, _ = get_readonly_field_data(field_name, obj)
    return display_for_value(value.humanized_value if isinstance(value, ReadonlyValue) else value, request=request)


def display_for_value(value, request=None):
    """
    Converts humanized value

    examples:
        boolean True/Talse ==> Yes/No
        objects ==> object display name with link if current user has permissions to see the object
        datetime ==> in localized format
    """
    from is_core.utils.compatibility import admin_display_for_value

    if request and isinstance(value, Model):
        return render_model_object_with_link(request, value)
    else:
        return (
            (value and ugettext('Yes') or ugettext('No')) if isinstance(value, bool) else admin_display_for_value(value)
        )


def get_url_from_model_core(request, obj):
    """
    Returns object URL from model core.
    """
    from is_core.site import get_model_core
    model_core = get_model_core(obj.__class__)

    if model_core and hasattr(model_core, 'ui_patterns'):
        edit_pattern = model_core.ui_patterns.get('detail')
        return (
            edit_pattern.get_url_string(request, obj=obj)
            if edit_pattern and edit_pattern.has_permission('get', request, obj=obj) else None
        )
    else:
        return None


def get_obj_url(request, obj):
    """
    Returns object URL if current logged user has permissions to see the object
    """
    if (is_callable(getattr(obj, 'get_absolute_url', None)) and
            (not hasattr(obj, 'can_see_edit_link') or
             (is_callable(getattr(obj, 'can_see_edit_link', None)) and obj.can_see_edit_link(request)))):
        return call_method_with_unknown_input(obj.get_absolute_url, request=request)
    else:
        return get_url_from_model_core(request, obj)


def render_model_object_with_link(request, obj, display_value=None):
    obj_url = get_obj_url(request, obj)
    display_value = force_text(obj) if display_value is None else display_value
    return format_html('<a href="{}">{}</a>', obj_url, display_value) if obj_url else display_value


def header_name_to_django(header_name):
    return '_'.join(('HTTP', header_name.replace('-', '_').upper()))


def pretty_class_name(class_name):
    return re.sub(r'(\w)([A-Z])', r'\1-\2', class_name).lower()


def get_export_types_with_content_type(export_types):
    generated_export_types = []
    for title, type, serialization_format in export_types:
        try:
            generated_export_types.append(
                (title, type, serialization_format, get_converter(type).media_type)
            )
        except KeyError:
            raise ImproperlyConfigured('Missing converter for type {}'.format(type))
    return generated_export_types


def get_link_or_none(pattern_name, request, view_kwargs=None):
    """
    Helper that generate URL prom pattern name and kwargs and check if current request has permission to open the URL.
    If not None is returned.

    Args:
        pattern_name (str): slug which is used for view registratin to pattern
        request (django.http.request.HttpRequest): Django request object
        view_kwargs (dict): list of kwargs necessary for URL generator

    Returns:

    """
    from is_core.patterns import reverse_pattern

    pattern = reverse_pattern(pattern_name)
    assert pattern is not None, 'Invalid pattern name {}'.format(pattern_name)

    if pattern.has_permission('get', request, view_kwargs=view_kwargs):
        return pattern.get_url_string(request, view_kwargs=view_kwargs)
    else:
        return None


class GetMethodFieldMixin:

    def get_method_returning_field_value(self, field_name):
        """
        Method should return object method that can be used to get field value.
        Args:
            field_name: name of the field

        Returns: method for obtaining a field value

        """
        method = getattr(self, field_name, None)
        return method if method and callable(method) else None



import os
from distutils.version import StrictVersion
from itertools import chain

import django
from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.core.validators import EMPTY_VALUES
from django.db.models.base import Model
from django.db.models.fields.files import FieldFile
from django.forms.utils import flatatt
from django.forms.widgets import MultiWidget, TextInput, Widget
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.html import conditional_escape, format_html, format_html_join
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from is_core.config import settings
from is_core.utils.compatibility import CompatibilityWidgetMixin

from .utils import ReadonlyValue, add_class_name

EMPTY_VALUE = '---'


def flat_data_attrs(attrs):
    return format_html_join('', ' data-{0}="{1}"', sorted(attrs.items()))


class WrapperWidget(forms.Widget):

    def __init__(self, widget):
        self.widget = widget

    @property
    def media(self):
        return self.widget.media

    @property
    def attrs(self):
        return self.widget.attrs

    def build_attrs(self, *args, **kwargs):
        "Helper function for building an attribute dictionary."
        self.attrs = self.widget.build_attrs(*args, **kwargs)
        return self.attrs

    def value_from_datadict(self, data, files, name):
        return self.widget.value_from_datadict(data, files, name)

    def id_for_label(self, id_):
        return self.widget.id_for_label(id_)

    def render(self, name, value, attrs=None):
        return self.widget.render(name, value, attrs)


class FulltextSelect(forms.Select):

    class_name = 'fulltext-search'
    placeholder = _('Search...')


class FulltextSelectMultiple(forms.SelectMultiple):

    class_name = 'fulltext-search-multiple'
    placeholder = _('Search...')


class ClearableFileInput(forms.ClearableFileInput):

    def render(self, name, value, attrs=None):
        if StrictVersion(django.get_version()) >= StrictVersion('1.8'):
            return super(ClearableFileInput, self).render(name, value, attrs)
        else:
            substitutions = {
                'initial_text': self.initial_text,
                'input_text': self.input_text,
                'clear_template': '',
                'clear_checkbox_label': self.clear_checkbox_label,
            }
            template = '%(input)s'
            substitutions['input'] = super(forms.ClearableFileInput, self).render(name, value, attrs)

            if value and hasattr(value, "url"):
                template = self.template_with_initial
                substitutions['initial'] = format_html(self.url_markup_template,
                                                       value.url,
                                                       os.path.basename(force_text(value)))
                if not self.is_required:
                    checkbox_name = self.clear_checkbox_name(name)
                    checkbox_id = self.clear_checkbox_id(checkbox_name)
                    substitutions['clear_checkbox_name'] = conditional_escape(checkbox_name)
                    substitutions['clear_checkbox_id'] = conditional_escape(checkbox_id)
                    substitutions['clear'] = forms.CheckboxInput().render(checkbox_name, False, attrs={'id': checkbox_id})
                    substitutions['clear_template'] = self.template_with_clear % substitutions

            return mark_safe(template % substitutions)

    def get_template_substitution_values(self, value):
        """
        Return value-related substitutions.
        """
        return {
            'initial': os.path.basename(conditional_escape(value)),
            'initial_url': conditional_escape(value.url),
        }


class DragAndDropFileInput(ClearableFileInput):

    def _render_value(self, value):
        return '<a href="%s">%s</a>' % (value.url, os.path.basename(value.name))

    def render(self, name, value, attrs={}):
        output = ['<div class="drag-and-drop-wrapper">']
        output.append('<div class="drag-and-drop-placeholder"%s></div>' % (id and 'data-for="%s"' % id or ''))
        output.append('<div class="thumbnail-wrapper">')
        if value and isinstance(value, FieldFile):
            output.append(self._render_value(value))
        output.append('</div><div class=file-input-wrapper>')
        output.append(super(DragAndDropFileInput, self).render(name, value, attrs=attrs))
        output.append(2 * '</div>')
        return mark_safe('\n'.join(output))


class SmartWidgetMixin:

    def smart_render(self, request, name, value, initial_value, form, *args, **kwargs):
        return self.render(name, value, *args, **kwargs)


class ReadonlyWidget(SmartWidgetMixin, Widget):

    def __init__(self, widget=None, attrs=None):
        super(ReadonlyWidget, self).__init__(attrs)
        self.widget = widget

    def _get_widget(self):
        widget = self.widget
        while isinstance(widget, WrapperWidget):
            widget = widget.widget
        return widget

    def _get_value(self, value):
        widget = self._get_widget()
        if widget and hasattr(widget, 'choices'):
            choices_dict = dict(widget.choices)
            text_choices_dict = {str(k): v for k, v in widget.choices}
            return choices_dict.get(value, text_choices_dict.get(value, value))
        else:
            return value

    def _get_value_display(self, value):
        from is_core.utils import display_for_value

        value = self._get_value(value)

        if value in EMPTY_VALUES:
            value = EMPTY_VALUE

        return display_for_value(value)

    def _render(self, name, value, attrs=None, choices=()):
        if isinstance(value, (list, tuple)):
            out = ', '.join([self._get_value_display(val) for val in value])
        else:
            out = self._get_value_display(value)
        return format_html('<p>{}</p>', out)

    def _smart_render(self, request, name, value, initial_value, form, attrs=None, choices=()):
        return self._render(name, value, attrs, choices)

    def _render_readonly_value(self, readonly_value):
        return mark_safe('<p><span title="%s">%s</span></p>' % (conditional_escape(force_text(readonly_value.value)),
                                                                conditional_escape(readonly_value.humanized_value)))

    def render(self, name, value, attrs=None, choices=()):
        if isinstance(value, ReadonlyValue):
            return self._render_readonly_value(value)
        else:
            return self._render(name, value, attrs, choices)

    def smart_render(self, request, name, value, initial_value, form, attrs=None, choices=()):
        return (
            self._render_readonly_value(value) if isinstance(value, ReadonlyValue)
            else self._smart_render(request, name, value, initial_value, form, attrs, choices)
        )

    def _has_changed(self, initial, data):
        return False


class ModelObjectReadonlyWidget(ReadonlyWidget):

    def _render_object(self, request, obj, display_value=None):
        from is_core.utils import render_model_object_with_link

        return render_model_object_with_link(request, obj, display_value)

    def _smart_render(self, request, name, value, initial_value, form, *args, **kwargs):
        if value and isinstance(value, Model):
            return format_html('<p>{}</p>', self._render_object(request, value))
        else:
            return super()._smart_render(request, name, value, initial_value, form, *args, **kwargs)


class NullBooleanReadonlyWidget(ReadonlyWidget):

    def _get_value(self, value):
        return {
            '2': True,
            True: True,
            'True': True,
            '3': False,
            'False': False,
            False: False,
        }.get(value)


class ManyToManyReadonlyWidget(ModelObjectReadonlyWidget):

    def _smart_render(self, request, name, value, initial_value, form, *args, **kwargs):
        if value and isinstance(value, (list, tuple)):
            return format_html_join(', ', '{}', ((self._render_object(request, obj),) for obj in value))
        else:
            return super(ModelObjectReadonlyWidget, self)._smart_render(request, name, value, initial_value, form,
                                                                        *args, **kwargs)


class ModelChoiceReadonlyWidget(ModelObjectReadonlyWidget):

    def _choice(self, value):
        widget = self._get_widget()
        if hasattr(widget, 'choices'):
            return widget.choices.get_choice_from_value(value)

    def _smart_render(self, request, name, value, initial_value, form, attrs=None, choices=()):
        choice = self._choice(value)

        if choice:
            value = self._render_object(request, choice.obj, force_text(choice[1]))
        elif value in EMPTY_VALUES:
            value = EMPTY_VALUE

        return format_html('<p>{}</p>', value)


class ModelMultipleReadonlyWidget(ModelChoiceReadonlyWidget):

    def _smart_render(self, request, name, values, initial_values, form, *args, **kwargs):
        if values and isinstance(values, (list, tuple)):
            rendered_values = []
            for value in values:
                choice = self._choice(value)
                if choice:
                    value = force_text(choice[1])
                    if choice.obj:
                        rendered_values.append(self._render_object(request, choice.obj, value))
                    else:
                        rendered_values.append(value)
            return format_html(
                '<p>{}</p>',
                format_html_join(', ', '{}', ((v,) for v in rendered_values)) if rendered_values else EMPTY_VALUE
            )
        else:
            return super(ModelObjectReadonlyWidget, self)._smart_render(request, name, values, initial_values, form,
                                                                        *args, **kwargs)


class URLReadonlyWidget(ReadonlyWidget):

    def _render(self, name, value, *args, **kwargs):
        if value:
            return format_html('<a href="{}">{}</a>', value, value)
        else:
            return super(URLReadonlyWidget, self)._render(name, value, *args, **kwargs)


class FileReadonlyWidget(ReadonlyWidget):

    def _render(self, name, value, *args, **kwargs):
        if value and isinstance(value, FieldFile):
            return format_html('<a href="{}">{}</a>', value.url, os.path.basename(value.name))
        else:
            return super()._render(name, value, *args, **kwargs)


class EmptyWidget(ReadonlyWidget):

    def render(self, *args, **kwargs):
        return ''

    def smart_render(self, *args, **kwargs):
        return ''


class ButtonWidget(CompatibilityWidgetMixin, ReadonlyWidget):

    def _render(self, name, value, attrs=None, *args, **kwargs):
        final_attrs = self.build_attrs(self.attrs, attrs, name=name)

        return format_html('<button %(attrs)s>%(value)s</button>' %
                           {'value': value, 'attrs': flatatt(final_attrs)})


class DivButtonWidget(CompatibilityWidgetMixin, ReadonlyWidget):

    def _render(self, name, value, attrs=None, *args, **kwargs):
        final_attrs = self.build_attrs(self.attrs, attrs)
        class_name = final_attrs.pop('class', '')
        return format_html('<div class="%(class_name)s btn btn-primary btn-small" '
                           '%(attrs)s>%(value)s</div>' %
                           {'class_name': class_name, 'value': value, 'attrs': flatatt(final_attrs)})


class MultipleTextInput(forms.TextInput):

    def __init__(self, attrs=None, separator=','):
        super(MultipleTextInput, self).__init__(attrs)
        self.separator = separator

    def render(self, name, value, attrs=None):
        if isinstance(value, str):
            value = [value]
        return super(MultipleTextInput, self).render(
            name, '{} '.format(self.separator).join(map(force_text, value)) if value else value, attrs
        )

    def value_from_datadict(self, data, files, name):
        value = super(MultipleTextInput, self).value_from_datadict(data, files, name)
        return [v.strip() for v in value.split(self.separator)] if isinstance(value, str) else value


class AbstractDateRangeWidget(MultiWidget):

    def __init__(self):
        class_from, class_to = self.get_range_classes()
        super(AbstractDateRangeWidget, self).__init__(
            (
                TextInput(attrs={'class': class_from}),
                TextInput(attrs={'class': class_to}),
            )
        )

    def get_range_classes(self):
        raise NotImplemented

    def decompress(self, value):
        return []


class DateRangeWidget(AbstractDateRangeWidget):

    def get_range_classes(self):
        return 'date-range-from', 'date-range-to'


class DateTimeRangeWidget(AbstractDateRangeWidget):

    def get_range_classes(self):
        return 'datetime-range-from', 'datetime-range-to'


class FilterDateRangeWidgetMixin:

    def render(self, name, value, attrs=None):
        if attrs and 'data-filter' in attrs:
            filter_term = attrs.pop('data-filter')
            for widget, operator in zip(self.widgets, ('gte', 'lt')):
                widget.attrs['data-filter'] = '{}__{}'.format(filter_term.rsplit('__', 1)[0], operator)
        return super(FilterDateRangeWidgetMixin, self).render(name, value, attrs)


class DateRangeFilterWidget(FilterDateRangeWidgetMixin, DateRangeWidget):
    pass


class DateTimeRangeFilterWidget(FilterDateRangeWidgetMixin, DateTimeRangeWidget):
    pass


class RestrictedSelectWidgetMixin(CompatibilityWidgetMixin):

    select_class_name = None
    select_class_placeholder = None

    @cached_property
    def is_restricted(self):
        """
        Returns True or False according to number of objects in queryset.
        If queryset contains too much objects the widget will be restricted and won't be used select box with choices.
        """
        return (
            not hasattr(self.choices, 'queryset') or
            self.choices.queryset.count() > settings.FOREIGN_KEY_MAX_SELECBOX_ENTRIES
        )

    def render(self, name, value, attrs=None):
        if self.is_restricted:
            if value is None:
                value = ''
            final_attrs = self.build_attrs(self.attrs, attrs, type='text', name=name)
            if value != '':
                # Only add the 'value' attribute if a value is non-empty.
                final_attrs['value'] = self.format_resticted_value(value)
            return format_html('<input{} />', flatatt(final_attrs))
        else:
            attrs = add_class_name(attrs, self.select_class_name)
            attrs['placeholder'] = self.select_class_placeholder
            return super(RestrictedSelectWidgetMixin, self).render(name, value, attrs=attrs)


class RestrictedSelectWidget(RestrictedSelectWidgetMixin, forms.Select):

    select_class_name = 'fulltext-search'
    select_class_placeholder = _('Search...')

    def format_resticted_value(self, value):
        return None if value is None else str(value)


class RestrictedSelectMultipleWidget(RestrictedSelectWidgetMixin, forms.SelectMultiple):

    select_class_name = 'fulltext-search-multiple'
    select_class_placeholder = _('Search...')

    def __init__(self, attrs=None, separator=','):
        super(RestrictedSelectMultipleWidget, self).__init__(attrs)
        self.separator = separator

    def value_from_datadict(self, data, files, name):
        if self.is_restricted:
            value = data.get(name)
            return [v.strip() for v in value.split(self.separator) if v.strip()] if isinstance(value, str) else value
        else:
            return super(RestrictedSelectMultipleWidget, self).value_from_datadict(data, files, name)

    def format_resticted_value(self, value):
        if value is None:
            value = []
        if not isinstance(value, (tuple, list)):
            value = [value]
        return self.separator.join([str(v) if v is not None else '' for v in value])



from django.db.models.query import QuerySet
from django.shortcuts import render
from django.utils import translation
from django.utils.encoding import force_text
from django.utils.translation import ugettext

from is_core.config import settings
from is_core.rest.resource import RESTResource, UIRESTModelResource

from security.tasks import obj_to_string

from .tasks import background_serialization


class ErrorResponseData(dict):

    def __init__(self, msg, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self['messages'] = {'error': force_text(msg)}


class CeleryResourceMixin:

    def _get_error_response_data(self, message):
        return {
            'messages': {'error': force_text(message)}
        }

    def _get_too_large_response_data(self, http_headers):
        return ErrorResponseData(
            ugettext('Too much data, please use filter or pagination to reduce its amount.')
        ), http_headers, 430, False

    def _get_no_background_permissions_response_data(self, http_headers):
        return ErrorResponseData(
            ugettext('User doesn\'t have permissions to export')
        ), http_headers, 403, False

    def _serialize_data_in_background(self, result):
        background_serialization.apply_async_on_commit(
            args=(
                self.request.user.pk,
                self.request._rest_context,
                translation.get_language(),
                force_text(self._get_requested_fieldset(result)),
                self._get_serialization_format(),
                self._get_filename(),
                obj_to_string(result.query),
            ),
            queue=settings.BACKGROUND_EXPORT_TASK_QUEUE
        )

    def _get_filename(self):
        parts = super()._get_filename().rsplit(sep='.', maxsplit=1)
        return ('{}{}.{}'.format(parts[0], '_DUVERNE', parts[1]) if len(parts) == 2 else
                '{}{}'.format(super()._get_filename(), '_DUVERNE')
        )

    def _get_background_serialization_response_data(self, result, http_headers):
        if not self.permission.has_permission('export', self.request, self):
            return self._get_no_background_permissions_response_data(http_headers)
        else:
            self._serialize_data_in_background(result)
            response = render(self.request, 'is_core/background_serialization.html')
            return response, response['Content-Type'], 200, False

    def _is_excluded_serialization_limit(self, result):
        return (
            'background_serialization' not in self.request._rest_context and
            isinstance(result, QuerySet) and
            result.count() > settings.BACKGROUND_EXPORT_SERIALIZATION_LIMIT
        )

    def _get_response_data(self):
        result, http_headers, status_code, fieldset = super()._get_response_data()
        if self._is_excluded_serialization_limit(result):
            return self._get_too_large_response_data(http_headers)
        elif 'background_serialization' in self.request._rest_context:
            return self._get_background_serialization_response_data(result, http_headers)
        else:
            return result, http_headers, status_code, fieldset

    def _get_headers_queryset_context_mapping(self):
        context_mapping = super()._get_headers_queryset_context_mapping()
        context_mapping['background_serialization'] = ('BACKGROUND_SERIALIZATION', '_background_serialization')
        return context_mapping

    def _get_name(self):
        obj = self._get_obj_or_none(pk=self.kwargs.get('pk'))
        return force_text(obj).replace(' ', '-') if obj else super()._get_name()


class CeleryRESTModelResource(CeleryResourceMixin, UIRESTModelResource):
    pass


class CeleryRESTResource(CeleryResourceMixin, RESTResource):
    pass



from django.db.models import Model
from django.views.generic.base import TemplateView
from django.utils.translation import ugettext_lazy as _

from is_core.utils import display_object_data, display_for_value
from is_core.generic_views import DefaultCoreViewMixin


class DataRow(list):

    def __init__(self, values, class_names):
        super(DataRow, self).__init__(values)
        self.class_names = class_names


class ObjectsViewMixin:
    """
    This class behaves and displays like table bud data are reaonly. No need of any form or queryset.
    Implement method 'get_objects' and define a list of fields in the attribute 'fields'.
    """

    # Tuple of tuples. The tuple consists of key and value (which will be displayed in template)
    # For eg: (('company_name', _('Company name')), ('zip', _('ZIP code')), ('...', '...'))
    fields = ()
    obj_class_names = ()
    no_items_text = _('There are no items')

    def get_fields(self):
        return list(self.fields)

    def parse_object(self, obj):
        return obj

    def get_objects(self):
        """
        This method must return a queryset of models, dictionaries or any iterable objects.

        Dictionaries must have keys defined in the attribute 'fields'. For eg: {'company_name': 'ABC', 'zip': '44455'}

        Objects / models must have attributes defined in the attribute 'fields'. Or it may have a humanizing methods
        'get_attributename_humanized', for eg: method 'get_zip_hunanized'.
        """
        raise NotImplementedError

    def get_data_list(self, fields, objects):
        out = []
        for obj in objects:
            normalized_obj = self.parse_object(obj)
            out.append(DataRow([(field_name, self.get_data_object(field_name, normalized_obj))
                                for field_name, _ in self.get_fields()],
                               self.get_obj_class_names(obj)))
        return out

    def get_obj_class_names(self, obj):
        return list(self.obj_class_names)

    def get_data_object(self, field_name, obj):
        if isinstance(obj, Model):
            return display_object_data(obj, field_name, request=self.request)
        elif isinstance(obj, dict):
            return display_for_value(obj.get(field_name), request=self.request)
        else:
            raise NotImplementedError

    def get_header_list(self, fields):
        return self.get_fields()

    def get_class_names(self):
        return [self.__class__.__name__.lower(), ]

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)
        context_data.update({
            'data_list': self.get_data_list(self.get_fields(), self.get_objects()),
            'header_list': self.get_header_list(self.get_fields()),
            'class_names': self.get_class_names(),
            'no_items_text': self.no_items_text
        })
        return context_data


class ObjectsView(ObjectsViewMixin, DefaultCoreViewMixin, TemplateView):

    template_name = 'is_core/generic_views/objects.html'



from distutils.version import StrictVersion

import django

from django.forms.widgets import DateInput, DateTimeInput, TimeInput, Widget
from django.forms.fields import (ImageField, FileField, Field, URLField, MultipleChoiceField, ChoiceField,
                                 NullBooleanField)

from is_core.forms.utils import add_class_name
from is_core.forms.widgets import (DragAndDropFileInput, ReadonlyWidget, URLReadonlyWidget,
                                   FileReadonlyWidget, FulltextSelectMultiple, NullBooleanReadonlyWidget)


if StrictVersion(django.get_version()) < StrictVersion('1.11'):
    def build_attrs(self, extra_attrs=None, **kwargs):
        """ Helper function for building an attribute dictionary. """
        attrs = dict(self.attrs, **kwargs)
        if extra_attrs:
            attrs.update(extra_attrs)
        if self.class_name:
            attrs = add_class_name(attrs, self.class_name)
        if self.placeholder:
            attrs['placeholder'] = self.placeholder
        return attrs
else:
    def build_attrs(self, base_attrs, extra_attrs=None, **kwargs):
        """ Helper function for building an attribute dictionary. """
        attrs = dict(base_attrs, **kwargs)
        if extra_attrs:
            attrs.update(extra_attrs)
        if self.class_name:
            attrs = add_class_name(attrs, self.class_name)
        if self.placeholder:
            attrs['placeholder'] = self.placeholder
        return attrs



Widget.placeholder = None
Widget.class_name = None
Widget.build_attrs = build_attrs

DateInput.class_name = 'date'
TimeInput.class_name = 'time'
DateTimeInput.class_name = 'datetime'

FileField.widget = DragAndDropFileInput
ImageField.widget = DragAndDropFileInput
MultipleChoiceField.widget = FulltextSelectMultiple

Field.is_readonly = False
Field.readonly_widget = ReadonlyWidget
URLField.readonly_widget = URLReadonlyWidget
FileField.readonly_widget = FileReadonlyWidget
NullBooleanField.readonly_widget = NullBooleanReadonlyWidget



from django.core.management.base import BaseCommand

from is_core.contrib.background_export.models import ExportedFile


class Command(BaseCommand):

    def handle(self, **options):
        expired_exported_files_qs = ExportedFile.objects.filter_expired()
        count_expired_exported_files = expired_exported_files_qs.count()

        if not count_expired_exported_files:
            self.stdout.write('No expired exported files were found.')
        else:
            [exported_file.remove_file() for exported_file in ExportedFile.objects.filter_expired()]
            self.stdout.write('{} expired exported files was removed.'.format(count_expired_exported_files))



from django.utils.translation import ugettext_lazy as _

from pyston.utils import JsonObj


class Action(JsonObj):

    def __init__(self, name, verbose_name, type, class_name=None):
        super(Action, self).__init__()
        self.name = name
        self.verbose_name = verbose_name
        self.type = type
        if class_name:
            self.class_name = class_name


class WebAction(Action):

    def __init__(self, name, verbose_name, class_name=None, target=None, rel=None):
        super(WebAction, self).__init__(name, verbose_name, 'web', class_name)
        if target:
            self.target = target
        if rel:
            self.rel = rel


class RESTAction(Action):

    def __init__(self, name, verbose_name, method, data=None, class_name=None, success_text=None):
        super(RESTAction, self).__init__(name, verbose_name, 'rest', class_name)
        self.method = method
        if data:
            self.data = data
        if success_text:
            self.success_text = success_text


class ConfirmRESTAction(RESTAction):

    def __init__(self, name, verbose_name, method, data=None, class_name=None, confirm_dialog=None, success_text=None):
        super(ConfirmRESTAction, self).__init__(name, verbose_name, method, data, class_name, success_text)
        self.confirm = confirm_dialog

    class ConfirmDialog(JsonObj):

        def __init__(self, text, title=None, true_label=None, false_label=None):
            self.true_label = true_label or _('Yes')
            self.false_label = false_label or _('No')
            self.title = title or _('Are you sure?')
            self.text = text



from django.forms.formsets import DEFAULT_MAX_NUM, BaseFormSet as OriginBaseFormSet


class BaseFormSetMixin:
    """
    Mixin that add method all_forms which return existing forms and empty form in one list
    """

    def __init__(self, *args, **kwargs):
        super(BaseFormSetMixin, self).__init__(*args, **kwargs)
        self._all_forms = None

    def all_forms(self):
        if self._all_forms is None:
            self._all_forms = []
            for form in self.forms:
                self._all_forms.append(form)

            if self.can_add:
                self._all_forms.append(self.empty_form)
        return self._all_forms

    def total_form_count(self):
        total_form_count = super(BaseFormSetMixin, self).total_form_count()
        if total_form_count < self.min_num:
            return self.min_num
        return total_form_count


class BaseFormSet(BaseFormSetMixin, OriginBaseFormSet):
    pass


def smartformset_factory(form, formset=BaseFormSet, extra=1, can_order=False,
                         can_delete=False, min_num=None, max_num=None, validate_min=False, validate_max=False):
    """Return a FormSet for the given form class."""
    if max_num is None:
        max_num = DEFAULT_MAX_NUM
    # hard limit on forms instantiated, to prevent memory-exhaustion attacks
    # limit is simply max_num + DEFAULT_MAX_NUM (which is 2*DEFAULT_MAX_NUM
    # if max_num is None in the first place)
    absolute_max = max_num + DEFAULT_MAX_NUM
    if min_num is None:
        min_num = 0

    attrs = {'form': form, 'extra': extra,
             'can_order': can_order, 'can_delete': can_delete, 'min_num': min_num,
             'max_num': max_num, 'absolute_max': absolute_max, 'validate_min': validate_min,
             'validate_max': validate_max}
    return type(form.__name__ + str('FormSet'), (formset,), attrs)



import os
from datetime import timedelta
from uuid import uuid4 as uuid

import import_string

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.shortcuts import resolve_url
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _

from is_core.config import settings as is_core_settings
from is_core.utils.decorators import short_description

from chamber.models import SmartModel
from chamber.models.fields import FileField


if 'security' not in settings.INSTALLED_APPS:
    raise ImproperlyConfigured('Missing library security, please install it.')


def generate_filename(instance, filename):
    return os.path.join('exports', instance.slug, filename)


class ExportedFileManager(models.QuerySet):

    def filter_expired(self):
        return self.filter_active(
            created_at__lt=timezone.now() - timedelta(days=is_core_settings.BACKGROUND_EXPORT_EXPIRATION_DAYS)
        )

    def filter_active(self, *args, **kwargs):
        return self.exclude(file='').filter(*args, **kwargs)


storage = import_string(is_core_settings.BACKGROUND_EXPORT_STORAGE_CLASS)()


class ExportedFile(SmartModel):

    objects = ExportedFileManager.as_manager()

    task = models.OneToOneField(
        'security.CeleryTaskLog',
        verbose_name=_('state'),
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    slug = models.SlugField(
        verbose_name=_('slug'),
        null=False,
        blank=False,
        primary_key=True,
        max_length=32
    )
    file = FileField(
        verbose_name=_('file'),
        null=True,
        blank=True,
        upload_to=generate_filename,
        max_upload_size=100,
        storage=storage
    )
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_('created by'),
        null=False,
        blank=False,
        related_name='created_exported_files',
        on_delete=models.PROTECT,
    )
    downloaded_by = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        verbose_name=_('downloaded by'),
        blank=True,
        related_name='downloaded_exported_files'
    )
    content_type = models.ForeignKey(
        ContentType,
        null=False,
        blank=False,
        on_delete=models.PROTECT,
    )

    @property
    def download_url(self):
        from security.enums import CeleryTaskLogState

        return (
            resolve_url('pyston-download-export', slug=self.slug)
            if self.file and self.task and self.task.state == CeleryTaskLogState.SUCCEEDED
            else ''
        )

    @short_description(_('download'))
    def download_link(self):
        return (
            format_html('<a href="{}">{}</a>', self.download_url, os.path.basename(self.file.name))
            if self.download_url else ''
        )

    def generate_slug(self):
        self.slug = uuid().hex

    def clean_slug(self):
        if not self.slug:
            self.generate_slug()

    def remove_file(self, *args, **kwargs):
        if self.file:
            directory_path = self.file.storage.path(os.path.join(self.file.path, '..'))
            if os.path.exists(directory_path):
                self.file.storage.delete(self.file.path)
                os.rmdir(directory_path)
            self.file = None
            self.save()

    @short_description(_('expiration'))
    def expiration(self):
        return (
            self.created_at + timedelta(days=is_core_settings.BACKGROUND_EXPORT_EXPIRATION_DAYS)
            if self.created_at else None
        )

    def __str__(self):
        return os.path.basename(self.file.name)

    class Meta:
        verbose_name = _('exported file')
        verbose_name_plural = _('exported files')
        ordering = ('-created_at',)



import re
import warnings
import datetime

from collections import OrderedDict

import django
from django.forms.forms import DeclarativeFieldsMetaclass, Form
from django.forms.fields import FileField
from django.core.exceptions import ValidationError
from django.utils.encoding import force_text
from django.utils.functional import cached_property

from pyston.forms import RESTFormMixin

from .fields import SmartReadonlyField
from .widgets import SmartWidgetMixin
from .boundfield import ReadonlyBoundField, SmartBoundField


class SmartFormMetaclass(DeclarativeFieldsMetaclass):

    def __new__(cls, name, bases, attrs):
        new_class = super(SmartFormMetaclass, cls).__new__(cls, name, bases, attrs)

        base_readonly_fields = set(getattr(new_class, 'base_readonly_fields', ()))
        base_required_fields = set(getattr(new_class, 'base_required_fields', ()))

        opts = getattr(new_class, 'Meta', None)
        if opts:
            setattr(opts, 'fields_to_clean', getattr(opts, 'fields_to_clean', None))
            exclude_fields = getattr(opts, 'exclude', None) or ()
            readonly_fields = getattr(opts, 'readonly_fields', None) or ()
            readonly = getattr(opts, 'readonly', None) or False
            required_fields = getattr(opts, 'required_fields', None) or ()

            for field_name, field in new_class.base_fields.items():
                if field_name in exclude_fields:
                    del new_class.base_fields[field_name]
                elif ((field_name in readonly_fields or readonly or field.is_readonly) and
                      field_name not in base_readonly_fields):
                    base_readonly_fields.add(field_name)
                elif field_name in required_fields and field_name not in base_required_fields:
                    base_required_fields.add(field_name)

            for field_name in set(readonly_fields):
                if (field_name not in new_class.base_fields and 'formreadonlyfield_callback' in attrs and
                        attrs['formreadonlyfield_callback'] is not None):
                    new_class.base_fields[field_name] = attrs['formreadonlyfield_callback'](field_name)
                    base_readonly_fields.add(field_name)

        new_class.base_readonly_fields = base_readonly_fields
        new_class.base_required_fields = base_required_fields
        return new_class


class SmartFormMixin:

    def __init__(self, *args, **kwargs):
        super(SmartFormMixin, self).__init__(*args, **kwargs)
        self.readonly_fields = set(self.base_readonly_fields)
        self._pre_init_fields()
        for field_name, field in self.fields.items():
            if hasattr(self, '_init_{}'.format(field_name)):
                getattr(self, '_init_{}'.format(field_name))(field)
        self._init_fields()

    def _pre_init_fields(self):
        for required_field_name in self.base_required_fields:
            if required_field_name in self.fields:
                self.fields[required_field_name].required = True

    def _init_fields(self):
        pass

    def __getitem__(self, name):
        """"
        Returns a BoundField with the given name.
        """
        try:
            field = self.fields[name]
        except KeyError:
            raise KeyError(
                "Key '%s' not found in '%s'. Choices are: %s." % (
                    name,
                    self.__class__.__name__,
                    ', '.join(sorted(f for f in self.fields)),
                )
            )
        if name not in self._bound_fields_cache:
            self._bound_fields_cache[name] = self._get_bound_field(name, field)
        return self._bound_fields_cache[name]

    def _get_bound_field(self, name, field):
        if name in self.readonly_fields:
            return ReadonlyBoundField(self, field, name)
        else:
            return SmartBoundField(self, field, name)

    def _get_readonly_widget(self, field_name, field, widget):
        if field.readonly_widget:
            if isinstance(field.readonly_widget, type):
                return field.readonly_widget(widget)
            else:
                return field.readonly_widget
        return field.widget

    def _get_filtered_fields_to_clean(self):
        return ((name, field) for name, field in self.fields.items()
                if self._get_fields_to_clean() is None or name in self._get_fields_to_clean())

    def _get_fields_to_clean(self):
        return self.Meta.fields_to_clean if hasattr(self, 'Meta') else None

    def _clean_fields(self):
        tmp_fields = self.fields
        self.fields = OrderedDict((
            (name, field) for name, field in self._get_filtered_fields_to_clean()
            if name not in self.readonly_fields
        ))
        super(SmartFormMixin, self)._clean_fields()
        self.fields = tmp_fields

    def _register_readonly_field(self, field_name):
        if field_name not in self.readonly_fields:
            self.readonly_fields.append(field_name)

    def _unregister_readonly_field(self, field_name):
        if field_name in self.readonly_fields:
            self.readonly_fields.remove(field_name)

    @cached_property
    def changed_data(self):
        tmp_fields = self.fields
        self.fields = OrderedDict((
            (name, field) for name, field in self.fields.items()
            if name not in self.readonly_fields
        ))
        changed_data = super(SmartFormMixin, self).changed_data
        self.fields = tmp_fields
        return changed_data


class SmartForm(SmartFormMixin, RESTFormMixin, Form, metaclass=SmartFormMetaclass):
    pass


def smartform_factory(request, form, readonly_fields=None, required_fields=None, exclude=None,
                      formreadonlyfield_callback=None, fields_to_clean=None, readonly=False):
    attrs = {}
    if exclude is not None:
        attrs['exclude'] = exclude
    if readonly_fields is not None:
        attrs['readonly_fields'] = readonly_fields
    if required_fields is not None:
        attrs['required_fields'] = required_fields
    if fields_to_clean is not None:
        attrs['fields_to_clean'] = fields_to_clean
    attrs['readonly'] = readonly
    # If parent form class already has an inner Meta, the Meta we're
    # creating needs to inherit from the parent's inner meta.
    parent = (object,)
    if hasattr(form, 'Meta'):
        parent = (form.Meta, object)
    Meta = type(str('Meta'), parent, attrs)

    class_name = form.__name__

    form_class_attrs = {
        'Meta': Meta,
        'formreadonlyfield_callback': formreadonlyfield_callback,
        '_request': request,
    }

    form_class = type(form)(class_name, (form,), form_class_attrs)
    return form_class



import re
import logging

from collections import OrderedDict

from django.conf.urls import url

from is_core.utils import get_new_class_name
from is_core.utils.compatibility import reverse, resolve


logger = logging.getLogger('is-core')

patterns = {}


def reverse_ui_view(name, request):
    pattern = patterns.get(name)
    if isinstance(pattern, UIPattern):
        return pattern.get_view(request)


def reverse_pattern(name):
    return patterns.get(name)


def pattern_from_request(request):
    return reverse_pattern(resolve(request.path).url_name)


class Pattern:

    send_in_rest = True

    def __init__(self, name):
        self.name = name
        self.url_prefix = None
        self._register()

    def _register(self):
        if self.name in patterns:
            logger.warning('Pattern with name %s has been registered yet' % self.name)
        else:
            patterns[self.name] = self

    def get_url_string(self, request, obj=None, view_kwargs=None):
        raise NotImplemented

    def get_url(self):
        return None


class ViewPattern(Pattern):

    def __init__(self, name, site_name, url_pattern, core):
        super(ViewPattern, self).__init__(name)
        self.url_pattern = url_pattern
        self.site_name = site_name
        self.core = core
        self.url_prefix = self.get_url_prefix()

    def get_url_prefix(self):
        if self.core:
            return self.core.get_url_prefix()

    @property
    def pattern(self):
        return '%s:%s' % (self.site_name, self.name)

    def _get_try_kwargs(self, request, obj):
        if obj and ('(?P<pk>[-\w]+)' in self.url_pattern or '(?P<pk>\d+)' in self.url_pattern):
            return {'pk': obj.pk}
        return {}

    def get_url_string(self, request, obj=None, view_kwargs=None):
        view_kwargs = {} if view_kwargs is None else view_kwargs
        try_kwargs = self._get_try_kwargs(request, obj)
        try_kwargs.update(view_kwargs)
        return reverse(self.pattern, kwargs=try_kwargs)

    def get_view_dispatch(self):
        raise NotImplementedError

    def get_view(self, request, args=None, kwargs=None):
        raise NotImplementedError

    def get_url(self):
        url_pattern = self.url_pattern
        if self.url_prefix:
            url_pattern = self.url_pattern
            if url_pattern.startswith('^'):
                url_pattern = url_pattern[1:]
            url_pattern = '%s/%s' % (self.url_prefix, url_pattern)
        return url('^%s$' % url_pattern, self.get_view_dispatch(), name=self.name)

    def _get_called_permission_kwargs(self, request, obj):
        kwargs = {}
        if obj:
            kwargs['obj'] = obj
        return kwargs

    def _call_view_method_with_request(self, method_name, request, request_kwargs=None,
                                       method_args=None, method_kwargs=None, obj=None):
        request_kwargs = request_kwargs if request_kwargs is not None else {}
        method_args = method_args if method_args is not None else ()
        method_kwargs = method_kwargs if method_kwargs is not None else {}

        bckp_request_kwargs = request.kwargs

        # The request kwargs must be always returned back
        try:
            try_request_kwargs = self._get_try_kwargs(request, obj)
            try_request_kwargs.update(request_kwargs)
            request.kwargs = try_request_kwargs
            view = self.get_view(request, None, try_request_kwargs)
            result = getattr(view, method_name)(*method_args, **method_kwargs)
        finally:
            request.kwargs = bckp_request_kwargs
        return result

    def has_permission(self, name, request, obj=None, view_kwargs=None):
        method_kwargs = self._get_called_permission_kwargs(request, obj)

        return self._call_view_method_with_request(
            'has_permission', request, request_kwargs=view_kwargs,
            method_args=(name,), method_kwargs=method_kwargs, obj=obj
        )


class UIPattern(ViewPattern):

    def __init__(self, name, site_name, url_pattern, view_class, core=None):
        super(UIPattern, self).__init__(name, site_name, url_pattern, core)
        self.view_class = type(str(get_new_class_name(name, view_class)), (view_class,), {})
        if core:
            self.view_class.__init_core__(core, self)

    def get_view(self, request, args=None, kwargs=None):
        view = self.view_class()
        if args:
            view.args = args
        if kwargs:
            view.kwargs = kwargs
        view.request = request
        return view

    def get_view_dispatch(self):
        return self.view_class.as_view()


class RESTPattern(ViewPattern):

    def __init__(self, name, site_name, url_pattern, resource_class, core=None, methods=None, clone_view_class=True):
        super(RESTPattern, self).__init__(name, site_name, url_pattern, core)
        if clone_view_class:
            self.resource_class = type(str(get_new_class_name(name, resource_class)), (resource_class,), {})
        else:
            self.resource_class = resource_class
        self.methods = methods
        if core:
            self.resource_class.__init_core__(core, self)

    def get_url_prefix(self):
        url_prefix = super(RESTPattern, self).get_url_prefix()
        if url_prefix:
            return 'api/{}'.format(url_prefix)

    def get_view(self, request, args=None, kwargs=None):
        view = self.resource_class(request)
        if args:
            view.args = args
        if kwargs:
            view.kwargs = kwargs
        return view

    def get_view_dispatch(self):
        return self.resource_class.as_view(allowed_methods=self.methods)

    def get_allowed_methods(self, request, obj):
        return self._call_view_method_with_request('check_permissions_and_get_allowed_methods', request,
                                                   method_kwargs={'restricted_methods': self.methods, 'obj': obj},
                                                   obj=obj)


class HiddenPatternMixin:
    send_in_rest = False


class HiddenRESTPattern(HiddenPatternMixin, RESTPattern):
    pass


class HiddenUIPattern(HiddenPatternMixin, UIPattern):
    pass


class DoubleRESTPattern:

    def __init__(self, resource_class, pattern_class, core, methods=None):
        self.resource_class = resource_class
        self.pattern_class = pattern_class
        self.core = core
        self.methods = set(methods) if methods else methods

    @property
    def patterns(self):
        detail_resource_methods = {'get', 'put', 'patch', 'delete', 'head', 'options'}
        list_resource_methods = self._get_list_allowed_methods()
        if self.methods is not None:
            detail_resource_methods &= self.methods
            list_resource_methods &= self.methods

        result = OrderedDict()
        if detail_resource_methods:
            result['api-resource'] = self.pattern_class(
                'api-resource-%s' % self.core.get_menu_group_pattern_name(), self.core.site_name, r'^(?P<pk>[-\w]+)/$',
                self.resource_class, self.core, detail_resource_methods, clone_view_class=False
            )
        if list_resource_methods:
            result['api'] = self.pattern_class(
                'api-%s' % self.core.get_menu_group_pattern_name(), self.core.site_name, r'$', self.resource_class,
                self.core, list_resource_methods, clone_view_class=False
            )
        return result

    def _get_list_allowed_methods(self):
        return (
            {'get', 'post', 'head', 'options', 'put'}
            if hasattr(self.core, 'is_bulk_change_enabled') and self.core.is_bulk_change_enabled()
            else {'get', 'post', 'head', 'options'}
        )
